#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
import os, sys, gtk, random

############################################################################################################################################################
########################################################## START gimpfu3 by J.F.Garcia [arakne] ############################################################
############################################################################################################################################################
#   Gimp-Python - allows the writing of GIMP plug-ins in Python.
#   Copyright (C) 1997  James Henstridge <james@daa.com.au>
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation; either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.

from gimpshelf import shelf
import string as _string
import math
import gimp
import gimpcolor
import re
from gimpenums import *
pdb = gimp.pdb

import gettext
t = gettext.translation("gimp20-python", gimp.locale_directory, fallback=True)
_ = t.ugettext

class error(RuntimeError): pass
class CancelError(RuntimeError): pass

PF_INT8        = PDB_INT8
PF_INT16       = PDB_INT16
PF_INT32       = PDB_INT32
PF_INT         = PF_INT32
PF_FLOAT       = PDB_FLOAT
PF_STRING      = PDB_STRING
PF_VALUE       = PF_STRING
PF_COLOR       = PDB_COLOR
PF_COLOUR      = PF_COLOR
PF_ITEM        = PDB_ITEM
PF_DISPLAY     = PDB_DISPLAY
PF_IMAGE       = PDB_IMAGE
PF_LAYER       = PDB_LAYER
PF_CHANNEL     = PDB_CHANNEL
PF_DRAWABLE    = PDB_DRAWABLE
PF_VECTORS     = PDB_VECTORS

PF_TOGGLE      = 1000
PF_BOOL        = PF_TOGGLE
PF_SLIDER      = 1001
PF_SPINNER     = 1002
PF_ADJUSTMENT  = PF_SPINNER

PF_FONT        = 1003
PF_FILE        = 1004
PF_BRUSH       = 1005
PF_PATTERN     = 1006
PF_GRADIENT    = 1007
PF_RADIO       = 1008
PF_TEXT        = 1009
PF_PALETTE     = 1010
PF_FILENAME    = 1011
PF_DIRNAME     = 1012
PF_OPTION      = 1013
PF_TAB         = 1014
PF_DRAW        = 1015
PF_TITLE       = 1016

_type_mapping = {
    PF_INT8        : PDB_INT8,
    PF_INT16       : PDB_INT16,
    PF_INT32       : PDB_INT32,
    PF_FLOAT       : PDB_FLOAT,
    PF_STRING      : PDB_STRING,
    PF_COLOR       : PDB_COLOR,
    PF_ITEM        : PDB_ITEM,
    PF_DISPLAY     : PDB_DISPLAY,
    PF_IMAGE       : PDB_IMAGE,
    PF_LAYER       : PDB_LAYER,
    PF_CHANNEL     : PDB_CHANNEL,
    PF_DRAWABLE    : PDB_DRAWABLE,
    PF_VECTORS     : PDB_VECTORS,

    PF_TOGGLE      : PDB_INT32,
    PF_SLIDER      : PDB_FLOAT,
    PF_SPINNER     : PDB_INT32,

    PF_FONT        : PDB_STRING,
    PF_FILE        : PDB_STRING,
    PF_BRUSH       : PDB_STRING,
    PF_PATTERN     : PDB_STRING,
    PF_GRADIENT    : PDB_STRING,
    PF_RADIO       : PDB_STRING,
    PF_TEXT        : PDB_STRING,
    PF_PALETTE     : PDB_STRING,
    PF_FILENAME    : PDB_STRING,
    PF_DIRNAME     : PDB_STRING,
    PF_OPTION      : PDB_INT32,
    PF_TAB         : PDB_STRING,
    PF_DRAW        : PDB_INT32,
    PF_TITLE       : PDB_STRING,
}

_obj_mapping = {
    PF_INT8        : int,
    PF_INT16       : int,
    PF_INT32       : int,
    PF_FLOAT       : float,
    PF_STRING      : str,
    PF_COLOR       : gimpcolor.RGB,
    PF_ITEM        : int,
    PF_DISPLAY     : gimp.Display,
    PF_IMAGE       : gimp.Image,
    PF_LAYER       : gimp.Layer,
    PF_CHANNEL     : gimp.Channel,
    PF_DRAWABLE    : gimp.Drawable,
    PF_VECTORS     : gimp.Vectors,

    PF_TOGGLE      : bool,
    PF_SLIDER      : float,
    PF_SPINNER     : int,

    PF_FONT        : str,
    PF_FILE        : str,
    PF_BRUSH       : str,
    PF_PATTERN     : str,
    PF_GRADIENT    : str,
    PF_RADIO       : str,
    PF_TEXT        : str,
    PF_PALETTE     : str,
    PF_FILENAME    : str,
    PF_DIRNAME     : str,
    PF_OPTION      : int,
    PF_TAB         : str,
    PF_DRAW        : int,
    PF_TITLE       : str,
}

_registered_plugins_ = {}

def register(proc_name, blurb, help, author, copyright, date, label, imagetypes, params, results, function, menu=None, domain=None, on_query=None, on_run=None):
    """This is called to register a new plug-in."""

    # First perform some sanity checks on the data
    def letterCheck(str):
        allowed = _string.letters + _string.digits + "_" + "-"
        for ch in str:
            if not ch in allowed:
		return 0
        else:
            return 1

    if not letterCheck(proc_name):
        raise error, "procedure name contains illegal characters"

    for ent in params:
        if len(ent) < 4:
            raise error, ("parameter definition must contain at least 4 "
                          "elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != int:
            raise error, "parameter types must be integers"

        if not letterCheck(ent[1]):
            raise error, "parameter name contains illegal characters"

    for ent in results:
        if len(ent) < 3:
            raise error, ("result definition must contain at least 3 elements (%s given: %s)" % (len(ent), ent))

        if type(ent[0]) != type(42):
            raise error, "result types must be integers"

        if not letterCheck(ent[1]):
            raise error, "result name contains illegal characters"

    plugin_type = PLUGIN

    if (not proc_name.startswith("python-") and
        not proc_name.startswith("python_") and
        not proc_name.startswith("extension-") and
        not proc_name.startswith("extension_") and
        not proc_name.startswith("plug-in-") and
        not proc_name.startswith("plug_in_") and
        not proc_name.startswith("file-") and
        not proc_name.startswith("file_")):
           proc_name = "python-fu-" + proc_name

    # if menu is not given, derive it from label
    need_compat_params = False
    if menu is None and label:
        fields = label.split("/")
        if fields:
            label = fields.pop()
            menu = "/".join(fields)
            need_compat_params = True

            import warnings
            message = ("%s: passing the full menu path for the menu label is "
                       "deprecated, use the 'menu' parameter instead" % (proc_name))
            warnings.warn(message, DeprecationWarning, 3)

        if need_compat_params and plugin_type == PLUGIN:
            file_params = [(PDB_STRING, "filename", "The name of the file", ""),
                           (PDB_STRING, "raw-filename", "The name of the file", "")]

            if menu is None:
                pass
            elif menu.startswith("<Load>"):
                params[0:0] = file_params
            elif menu.startswith("<Image>") or menu.startswith("<Save>"):
                params.insert(0, (PDB_IMAGE, "image", "Input image", None))
                params.insert(1, (PDB_DRAWABLE, "drawable", "Input drawable", None))
                if menu.startswith("<Save>"):
                    params[2:2] = file_params

    _registered_plugins_[proc_name] = (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run)

def _query():
    for plugin in _registered_plugins_.keys():
        (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[plugin]

        def make_params(params):
            return [(_type_mapping[x[0]], x[1], _string.replace(x[2], "_", "")) for x in params]

        params = make_params(params)
        # add the run mode argument ...
        params.insert(0, (PDB_INT32, "run-mode", "The run mode { RUN-INTERACTIVE (0), RUN-NONINTERACTIVE (1) }"))

        results = make_params(results)

        if domain:
            try:
                (domain, locale_dir) = domain
                gimp.domain_register(domain, locale_dir)
            except ValueError:
                gimp.domain_register(domain)

        gimp.install_procedure(plugin, blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results)
        if menu:
            gimp.menu_register(plugin, menu)
        if on_query:
            on_query()

def getXtra(xtra, valName, defValue):
    p = defValue
    if xtra.get(valName):
        p = xtra.get(valName)
    return p

def _get_defaults(proc_name):
    import gimpshelf

    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    key = "python-fu-save--" + proc_name

    if gimpshelf.shelf.has_key(key):
        return gimpshelf.shelf[key]
    else:
        # return the default values
        return [x[3] for x in params]

def _set_defaults(proc_name, defaults):
    import gimpshelf

    key = "python-fu-save--" + proc_name
    gimpshelf.shelf[key] = defaults

def _interact(proc_name, start_params):
    (blurb, help, author, copyright, date, label, imagetypes, plugin_type, params, results, function, menu, domain, on_query, on_run) = _registered_plugins_[proc_name]

    def run_script(run_params):
        params = start_params + tuple(run_params)
        _set_defaults(proc_name, params)
        return apply(function, params)

    params = params[len(start_params):]

    # short circuit for no parameters ...
    if len(params) == 0:
         return run_script([])

    import pygtk
    pygtk.require('2.0')

    import gimpui, gtk
    gimpui.gimp_ui_init ()  # add MrQ	
#    import pango

    defaults = _get_defaults(proc_name)
    defaults = defaults[len(start_params):]

    class EntryValueError(Exception):
        pass

    def warning_dialog(parent, primary, secondary=None):
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_WARNING, gtk.BUTTONS_CLOSE, primary)
        if secondary:
            dlg.format_secondary_text(secondary)
        dlg.run()
        dlg.destroy()

    def error_dialog(parent, proc_name):
        import sys, traceback

        exc_str = exc_only_str = _("Missing exception information")
        try:
            etype, value, tb = sys.exc_info()
            exc_str = "".join(traceback.format_exception(etype, value, tb))
            exc_only_str = "".join(traceback.format_exception_only(etype, value))
        finally:
            etype = value = tb = None

        title = _("An error occurred running %s") % proc_name
        dlg = gtk.MessageDialog(parent, gtk.DIALOG_DESTROY_WITH_PARENT, gtk.MESSAGE_ERROR, gtk.BUTTONS_CLOSE, title)
        dlg.format_secondary_text(exc_only_str)

        alignment = gtk.Alignment(0.0, 0.0, 1.0, 1.0)
        alignment.set_padding(0, 0, 12, 12)
        dlg.vbox.pack_start(alignment)
        alignment.show()

        expander = gtk.Expander(_("_More Information"));
        expander.set_use_underline(True)
        expander.set_spacing(6)
        alignment.add(expander)
        expander.show()

        scrolled = gtk.ScrolledWindow()
        scrolled.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
        scrolled.set_size_request(-1, 200)
        expander.add(scrolled)
        scrolled.show()

        label = gtk.Label(exc_str)
        label.set_alignment(0.0, 0.0)
        label.set_padding(6, 6)
        label.set_selectable(True)
        scrolled.add_with_viewport(label)
        label.show()

        def response(widget, id):
            widget.destroy()

        dlg.connect("response", response)
        dlg.set_resizable(True)
        dlg.show()

    # define a mapping of param types to edit objects ...
    class StringEntry(gtk.Entry):
        def __init__(self, default=""):
            gtk.Entry.__init__(self)
            self.set_text(str(default))
            #self.set_line_wrap(True)
            #self.set_single_line_mode(False)

        def get_value(self):
            return self.get_text()

    class LabelTitle(gtk.Label):
        def __init__(self, default="", align = (0.0, 0.0)):
            gtk.Label.__init__(self)
            self.set_label(str(default))
            self.set_use_markup(True)
            if align != None:
                self.set_alignment(align[0],align[1])
        def get_value(self):
            return self.get_label()

    class TextEntry(gtk.ScrolledWindow):
        def __init__ (self, default=""):
            gtk.ScrolledWindow.__init__(self)
            self.set_shadow_type(gtk.SHADOW_IN)

            self.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_AUTOMATIC)
            self.set_size_request(100, -1)

            self.view = gtk.TextView()
            self.add(self.view)
            self.view.show()
            self.buffer = self.view.get_buffer()
            self.set_value(str(default))

        def set_value(self, text):
            self.buffer.set_text(text)

        def get_value(self):
            return self.buffer.get_text(self.buffer.get_start_iter(), self.buffer.get_end_iter())

    class IntEntry(StringEntry):
        def get_value(self):
            try:
                return int(self.get_text())
            except ValueError, e:
                raise EntryValueError, e.args

    class FloatEntry(StringEntry):
            def get_value(self):
                try:
                    return float(self.get_text())
                except ValueError, e:
                    raise EntryValueError, e.args

    def precision(step):
        # calculate a reasonable precision from a given step size
        if math.fabs(step) >= 1.0 or step == 0.0:
            digits = 0
        else:
            digits = abs(math.floor(math.log10(math.fabs(step))));
        if digits > 20:
            digits = 20
        return int(digits)

    class SliderEntry(gtk.HScale):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.HScale.__init__(self, self.adj)
            self.set_digits(precision(step))

        def get_value(self):
            return self.adj.value

    class SpinnerEntry(gtk.SpinButton):
        # bounds is (upper, lower, step)
        def __init__(self, default=0, bounds=(0, 100, 5)):
            step = bounds[2]
            self.adj = gtk.Adjustment(default, bounds[0], bounds[1], step, 10 * step, 0)
            gtk.SpinButton.__init__(self, self.adj, step, precision(step))

    class ToggleEntry(gtk.ToggleButton):
        def __init__(self, default=0, labels=[_("No"),_("Yes")]):
            gtk.ToggleButton.__init__(self)
            self.labels = labels
            self.label = gtk.Label(labels[0])
            self.add(self.label)
            self.label.show()
            self.connect("toggled", self.changed)
            self.set_active(default)

        def changed(self, tog):
            if tog.get_active():
                self.label.set_text(self.labels[1])
            else:
                self.label.set_text(self.labels[0])

        def get_value(self):
            return self.get_active()

    class RadioEntry(gtk.VBox):
        def __init__(self, default=0, items=((_("Yes"), 1), (_("No"), 0))):
            gtk.VBox.__init__(self, homogeneous=False, spacing=2)

            button = None

            for (label, value) in items:
                button = gtk.RadioButton(button, label)
                self.pack_start(button)
                button.show()

                button.connect("toggled", self.changed, value)

                if value == default:
                    button.set_active(True)
                    self.active_value = value

        def changed(self, radio, value):
            if radio.get_active():
                self.active_value = value

        def get_value(self):
            return self.active_value

    class ComboEntry(gtk.ComboBox):
        def __init__(self, default=0, items=()):
            store = gtk.ListStore(str)
            for item in items:
                store.append([item])

            gtk.ComboBox.__init__(self, model=store)

            cell = gtk.CellRendererText()
            self.pack_start(cell)
            self.set_attributes(cell, text=0)

            self.set_active(default)

        def get_value(self):
            return self.get_active()

    def FileSelector(default=""):
        # FIXME: should this be os.path.separator?  If not, perhaps explain why?
        if default and default.endswith("/"):
            selector = DirnameSelector
            if default == "/": default = ""
        else:
            selector = FilenameSelector
        return selector(default)

    class FilenameSelector(gtk.FileChooserButton):
        def __init__(self, default="", save_mode=False):
            gtk.FileChooserButton.__init__(self,_("Python-Fu File Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_OPEN)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DirnameSelector(gtk.FileChooserButton):
        def __init__(self, default=""):
            gtk.FileChooserButton.__init__(self,_("Python-Fu Folder Selection"))
            self.set_action(gtk.FILE_CHOOSER_ACTION_SELECT_FOLDER)
            if default:
                self.set_filename(default)

        def get_value(self):
            return self.get_filename()

    class DrawCanvas(gtk.DrawingArea):
        def __init__(self,default, size=(100,100)):
            gtk.DrawingArea.__init__(self)
            self.set_size_request(size[0],size[1])
        def get_value(self):
            return 0

    _edit_mapping = {
            PF_INT8        : IntEntry,
            PF_INT16       : IntEntry,
            PF_INT32       : IntEntry,
            PF_FLOAT       : FloatEntry,
            PF_STRING      : StringEntry,
            PF_COLOR       : gimpui.ColorSelector,
            PF_ITEM        : IntEntry,  # should handle differently ...
            PF_IMAGE       : gimpui.ImageSelector,
            PF_LAYER       : gimpui.LayerSelector,
            PF_CHANNEL     : gimpui.ChannelSelector,
            PF_DRAWABLE    : gimpui.DrawableSelector,
            PF_VECTORS     : gimpui.VectorsSelector,

            PF_TOGGLE      : ToggleEntry,
            PF_SLIDER      : SliderEntry,
            PF_SPINNER     : SpinnerEntry,
            PF_RADIO       : RadioEntry,
            PF_OPTION      : ComboEntry,

            PF_FONT        : gimpui.FontSelector,
            PF_FILE        : FileSelector,
            PF_FILENAME    : FilenameSelector,
            PF_DIRNAME     : DirnameSelector,
            PF_BRUSH       : gimpui.BrushSelector,
            PF_PATTERN     : gimpui.PatternSelector,
            PF_GRADIENT    : gimpui.GradientSelector,
            PF_PALETTE     : gimpui.PaletteSelector,
            PF_TEXT        : TextEntry,
            PF_TAB         : TextEntry,
            PF_DRAW        : DrawCanvas,
            PF_TITLE       : LabelTitle,
    }
    # JFGARCIA - modified: to pass args to the on_run
    if on_run:
        on_run(start_params+tuple(params))

    dialog = gimpui.Dialog(proc_name, "python-fu", None, 0, None, proc_name, (gtk.STOCK_CANCEL, gtk.RESPONSE_CANCEL, gtk.STOCK_OK, gtk.RESPONSE_OK))

    dialog.set_alternative_button_order((gtk.RESPONSE_OK, gtk.RESPONSE_CANCEL))

    dialog.set_transient()

    vbox = gtk.VBox(False, 12)
    vbox.set_border_width(12)
    dialog.vbox.pack_start(vbox)
    vbox.show()

    if blurb:
        if domain:
            try:
                (domain, locale_dir) = domain
                trans = gettext.translation(domain, locale_dir, fallback=True)
            except ValueError:
                trans = gettext.translation(domain, fallback=True)
            blurb = trans.ugettext(blurb)
        box = gimpui.HintBox(blurb)
        vbox.pack_start(box, expand=False)
        box.show()

    table = gtk.Table(1, 2, False)
    table.set_row_spacings(2)
    table.set_col_spacings(2)
    vbox.pack_start(table, expand=True)
    table.show()

    notebook = gtk.Notebook()
    tables=[]
    for i in range(len(params)):
        if params[i][0]==PF_TAB:
            vals = params[i][4]
            currPage = params[i][3]

            for n in vals:
                tbl = gtk.Table(2, 2, False)
                tbl.set_row_spacings(5)
                tbl.set_col_spacings(5)
                tbl.show()
                tables.append({'table':tbl,'rows':0, 'totCols':0})
                notebook.append_page(tbl, gtk.Label(n))
                notebook.set_current_page(currPage)


    if len(tables)>0:
        table.attach(notebook, 0,6,0,1)
        notebook.show()
    else:
        tables.append({'table':table,'rows':0})

    def response(dlg, id):
        if id == gtk.RESPONSE_OK:
            dlg.set_response_sensitive(gtk.RESPONSE_OK, False)
            dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, False)
            params = []
            try:
                for wid in edit_wids:
                    params.append(wid.get_value())
            except EntryValueError:
                warning_dialog(dialog, _("Invalid input for '%s'") % wid.desc)
            else:
                try:
                    dialog.res = run_script(params)
                except Exception:
                    dlg.set_response_sensitive(gtk.RESPONSE_CANCEL, True)
                    error_dialog(dialog, proc_name)
                    raise
        gtk.main_quit()

    iniX=0
    dialog.connect("response", response)
    row=0
    edit_wids = []

    for i in range(len(params)):
        paramsi = params[i]
        pf_type = paramsi[0]
        name, desc = (paramsi[1], paramsi[2])
        def_val = defaults[i]
 
        tab, incr, rows, connect, cols, hlp = (0, 1, 1, None, 1, desc)
        if len(params[i])>5:
            xtra = params[i][5]
            incr = int(getXtra(xtra, 'newLine', 1))
            if incr==0 : cols = 1
            cols = getXtra(xtra, 'cols'   , cols)
            iniX = getXtra(xtra, 'iniCol' , iniX)
            rows = getXtra(xtra, 'rows'   , 1)
            tab  = getXtra(xtra, 'tab'    , 0)
            hlp  = getXtra(xtra, 'help'   , desc)
            connect = getXtra(xtra,'connect', None)
        else:
            iniX = 0
        tbl = tables[tab]['table']
        row = tables[tab]['rows']


        if pf_type==PF_TAB:
            wid = _edit_mapping[pf_type](None)
        else:
            if pf_type != PF_TITLE and desc != '':
                label = gtk.Label(desc)
                label.set_use_underline(True)
                label.set_alignment(0.0, 0.5)
                tbl.attach(label, iniX + 1, iniX + 2, row, row + rows, xoptions = gtk.FILL)
                label.show()

            if pf_type in (PF_SPINNER, PF_SLIDER, PF_RADIO, PF_OPTION, PF_DRAW ):
                wid = _edit_mapping[pf_type](def_val, paramsi[4])
            elif pf_type==PF_TOGGLE:
                if isinstance(paramsi[4], list):
                    wid = _edit_mapping[pf_type](def_val, paramsi[4])
                else:
                    wid = _edit_mapping[pf_type](def_val)
            elif pf_type==PF_TITLE:
                wid = _edit_mapping[pf_type](desc)
            else:
                wid = _edit_mapping[pf_type](def_val)
            if pf_type!=PF_TITLE and desc != '':
                label.set_mnemonic_widget(wid)
                tbl.attach(wid, iniX + 2, iniX + 2 + cols, row, row + rows, yoptions=0)
                iniX = iniX + cols + 1
            else:
                tbl.attach(wid, iniX, iniX+cols, row, row + rows, xoptions=gtk.FILL)
            # jfGarcia: strip html tags from toolTip
            tTip = re.sub('<[^<]+?>', '', hlp)
            if pf_type != PF_TEXT:
                wid.set_tooltip_text(tTip)
            else:
                # Attach tip to TextView, not to ScrolledWindow
                wid.view.set_tooltip_text(tTip)

            wid.set_name(name)
            if connect:
                try:
                    wid.connect(connect['signal'],connect['function'])
                except Exception, e:
                    pass
            wid.show()
            wid.desc = desc

        edit_wids.append(wid)

        tables[tab]['rows'] += incr
        if incr==1:
           iniX=0

    progress_vbox = gtk.VBox(False, 6)
    vbox.pack_end(progress_vbox, expand=False)
    progress_vbox.show()

    progress = gimpui.ProgressBar()
    progress_vbox.pack_start(progress)
    progress.show()

    dialog.show()

    gtk.main()

    if hasattr(dialog, "res"):
        res = dialog.res
        dialog.destroy()
        return res
    else:
        dialog.destroy()
        raise CancelError

def _run(proc_name, params):
    run_mode = params[0]
    func = _registered_plugins_[proc_name][10]

    if run_mode == RUN_NONINTERACTIVE:
        return apply(func, params[1:])

    script_params = _registered_plugins_[proc_name][8]

    min_args = 0
    if len(params) > 1:
        for i in range(1, len(params)):
            param_type = _obj_mapping[script_params[i - 1][0]]
            if not isinstance(params[i], param_type):
                break
        min_args = i
    if len(script_params) > min_args:
        start_params = params[:min_args + 1]
        if run_mode == RUN_WITH_LAST_VALS:
            default_params = _get_defaults(proc_name)
            params = start_params + default_params[min_args:]
        else:
            params = start_params
    else:
       run_mode = RUN_NONINTERACTIVE
    if run_mode == RUN_INTERACTIVE:
        try:
            res = _interact(proc_name, params[1:])
        except CancelError:
            return
    else:
        res = apply(func, params[1:])
        gimp.displays_flush()
    return res

def main():
    """This should be called after registering the plug-in."""
    gimp.main(None, None, _query, _run)

def fail(msg):
    """Display an error message and quit"""
    gimp.message(msg)
    raise error, msg

def N_(message):
    return message




############################################################################################################################################################
######################################################### END gimpfu3 ######################################################################################
############################################################################################################################################################
# -------------------------------------------------------------------------------------
# -------------------------------------------------------------------------------------
#
# -------------------------------------------------------------------------------------
#
# Copyright (c) 2022 Skinnyhouse
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification, 
# are permitted provided that the following conditions are met:
#
#    - Redistributions of source code must retain the above copyright notice, this 
#    list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above copyright notice, 
#    this list of conditions and the following disclaimer in the documentation and/or 
#    other materials provided with the distribution.
#    - Neither the name of the author nor the names of its contributors may be used 
#    to endorse or promote products derived from this software without specific prior 
#    written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY 
# EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
# OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT 
# SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
# TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR 
# BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN 
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN 
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH 
# DAMAGE.
#
# -------------------------------------------------------------------------------------
#
# This file is a basic Python plug-in for GIMP.
#
# It can be executed by selecting the menu option: 'Filters>Custom Font Tools>Custom Font Metallic Effects v0.03
# Version 0.01 applies a number of metallic effects to the character layers of a custom font
# Version 0.02 code optimized and option to apply metallics effects to the current active layer 
# Version 0.03 code further optimized, effects improved - especially when applying to original texture
# Version 0.04 now accepts unicode character layers
# Version 0.05 option to work on the active layer (applies effect to non-character layers)
# Version 0.06 Gimp 2.10.20 compatible (selective gaussian blur issue)
# Version 0.07 Added option requested by GJL to emulate a chrome effect scm filter by Denis Bodor
# Version 0.08 Added sets of curves and options to use these with plate texture
# Version 0.09 Fixed error applying filter to active layer instead of character layer
# Version 0.10 Added tabbed dialog courtesy of Arakne for creating Gimpui3 and MareroQ for making it available to me
#			   Re-ordered dialog options into two tabs
#			   Altered selective blur default value to make original texture option more easily accessible again
#              Added randomness to verdigris texture


# Some global colours we may need to define

black = gimpcolor.RGB(0, 0, 0)
white = gimpcolor.RGB(255, 255, 255)
darkestgrey = gimpcolor.RGB(20, 20, 20)
verydarkgrey = gimpcolor.RGB(50, 50, 50)
darkgrey = gimpcolor.RGB(75, 75, 75)
midgrey2 = gimpcolor.RGB(105, 105, 105)
midgrey = gimpcolor.RGB(125, 125, 125)
lightgrey = gimpcolor.RGB(170, 170, 170)
verylightgrey = gimpcolor.RGB(222, 222, 222)
red = gimpcolor.RGB(220,10,10)	

# Some global metallic colours we may need for creating gradients

lightsilver = gimpcolor.RGB(204, 202, 202)
darksilver = gimpcolor.RGB(46, 43, 43)
lightgold = gimpcolor.RGB(250, 233, 49)
darkgold = gimpcolor.RGB(240, 183, 42)
lightcopper = gimpcolor.RGB(212, 125, 1)
darkcopper = gimpcolor.RGB(177, 106, 58)
lightbrass = gimpcolor.RGB(253, 242, 91)
darkbrass = gimpcolor.RGB(220 , 118, 20)
lightbronze = gimpcolor.RGB(104, 75, 33)
darkbronze = gimpcolor.RGB(83, 69, 42)


def do_stuff(image, layer, metal_texture, metal_colour, metal_curves, plate_softness, plate_border, retain_original, texture_opacity, do_gradient, fill_gradient, effect_mode_ptr, bump_depth, final_bump_depth, 
    light_amount, compass, effect_elevation, do_sheen, apply_lighting, shade_of_grey, artifact_radius, artifact_delta, apply_to_custom, ptab):
    #### Main Filter Proc
	
    # Store the GIMP's settings so they can be restored
    gimp.context_push()
    # Set start undo point
    pdb.gimp_image_undo_group_start(image)	
    do_metal = True
    loop_ptr = 1
    first_char_layer_name = "none"
    num_layers, layer_ids = pdb.gimp_image_get_layers(image)
    loop = num_layers - loop_ptr
    layer_names_list = image.layers
    original_active_layer = image.active_layer
    offset_x,offset_y = pdb.gimp_drawable_offsets(original_active_layer)
    position_active = pdb.gimp_image_get_item_position(image, original_active_layer)
	
    gegl_blur = fn_check_version()
    if gegl_blur:
        #artifact_delta = float((artifact_delta/255))
        artifact_delta = round(artifact_delta,3)
        artifact_radius = round(float(artifact_radius),2)
    #pdb.gimp_message("artifact_delta = " + str(artifact_delta) + " artifact_radius = "  + str(artifact_radius) )
	
	
    if not (do_metal == False):
        # turn off visibility of layers and make a note of the character layer to set as visible when we have finished
        visi_loop, first_char_layer  = 0, -1
        multi_layer_list = "No non-character layers."
        pitted_drawable_name = "None"
        num_multi_layers,other_layers,num_char_layers = 0,0,0
        while visi_loop < num_layers:
	        # get info about layer we are working on
            layer_name = pdb.gimp_item_get_name(layer_names_list[visi_loop])
            layer_name = layer_name.decode('utf-8')
            test_for_custom_layer = fn_test_for_custom_layer(image, layer_names_list[visi_loop], layer_name)
			
            if test_for_custom_layer:
            #if len(layer_name) == 1 or len(layer_name) == 5 and ".png" in layer_name or len(layer_name) == 6 and ".png" in layer_name:
            #if layer_name<>"space.png":
                num_char_layers += 1
                if first_char_layer < 0: 
                    first_char_layer = visi_loop
                    first_char_layer_name = layer_name
                if apply_to_custom: layer_names_list[visi_loop].set_offsets(0, 0)
            else:
                pdb.gimp_image_select_item(image, 2, layer_names_list[visi_loop])
                is_empty = pdb.gimp_selection_is_empty(image)
                if is_empty == False:
                    # make sure that existing character layers are positioned at the origin of the image
                    if not apply_to_custom: layer_names_list[visi_loop].set_offsets(0, 0)
                    # make a list of the non-character layers
                    if other_layers == 0:                 
                        multi_layer_list = [layer_name]
                    else:                        
                        multi_layer_list.append(layer_name)
                    other_layers += 1
            pdb.gimp_item_set_visible(layer_names_list[visi_loop],False) # turn off visibility
            visi_loop += 1
			
        if other_layers>0: pitted_drawable_name = multi_layer_list[0] # set name of the layer we think can be used to copy from for verdigris texture option

    # check for verdigris texture layer and turn off use of original texture if no suitable layer is available
    if metal_texture == 2:
        if other_layers > 0:
            pitted_drawable = pdb.gimp_image_get_layer_by_name(image,pitted_drawable_name)
            position_pitted = pdb.gimp_image_get_item_position(image, pitted_drawable) # get position of verdigris layer in layer stack
            if position_active == position_pitted: # don't want to use the active layer as the verdigris layer in non-custom font mode
                not_pitted = True
                ploop = 0
                while ploop<len(multi_layer_list): # loop through layers and find a new layer to act as a verdigris layer if possible
                    pitted_layer = pdb.gimp_image_get_layer_by_name(image, multi_layer_list[ploop])
                    pdb.gimp_image_select_item(image, 2, pitted_layer)
                    empty_selection = pdb.gimp_selection_is_empty(image)
                    pdb.gimp_selection_none(image)
                    if ploop<>position_pitted and not empty_selection:
                        if not_pitted:
                            pitted_drawable_name = multi_layer_list[ploop]
                            not_pitted = False
                    ploop+= 1
            pitted_drawable = pdb.gimp_image_get_layer_by_name(image,pitted_drawable_name)
            position_pitted = pdb.gimp_image_get_item_position(image, pitted_drawable)
            if position_active == position_pitted: retain_original = False # still no suitable verdigris layer so turn off use of the original texture in non-custom font mode

    while loop >= 0 and apply_to_custom:
	    # get info about layer we are working on for custom font mode
        layer_id = layer_ids[loop]
        layer_name = layer_names_list[loop]
	
        # cater for unicode characters above normal range
        display_layer_name =  pdb.gimp_item_get_name(layer_names_list[loop])
        display_layer_name = display_layer_name.decode('utf-8')	
	
        # check we have the layer id of the currently active layer
        pdb.gimp_image_set_active_layer(image, layer_name)
        layer_id = pdb.gimp_image_get_active_layer(image)

        # check to see if the current layer is a custom font character layer
        test_for_custom_layer = fn_test_for_custom_layer(image, layer_id, display_layer_name)
		
		
        if test_for_custom_layer:

            # ensure we have the active layer and make it visible
            active_layer_id = pdb.gimp_image_get_active_layer(image)
            pdb.gimp_item_set_visible(active_layer_id, True)
			
            if do_metal:
                active_layer = image.active_layer
                active_layer_id = proc_metal(image, active_layer, shade_of_grey, fill_gradient, do_gradient, metal_colour, metal_texture, do_sheen, bump_depth, effect_elevation, effect_mode_ptr, retain_original, compass, light_amount, apply_lighting, artifact_radius, artifact_delta, other_layers, pitted_drawable_name, final_bump_depth, apply_to_custom, texture_opacity, metal_curves, plate_softness, plate_border)
                
				
				
            if (num_layers - loop_ptr) > 0: pdb.gimp_item_set_visible(active_layer_id, False)
			
            num_layers, layer_ids = pdb.gimp_image_get_layers(image)
			
        loop_ptr += 1
        loop = num_layers - loop_ptr
		
	
    if apply_to_custom and first_char_layer_name <> "none": # the character layer to make visible on completion
        first_char_layer_id = pdb.gimp_image_get_layer_by_name(image, first_char_layer_name)
        pdb.gimp_item_set_visible(first_char_layer_id, True)
        pdb.gimp_image_set_active_layer(image,first_char_layer_id)
		
    if apply_to_custom and first_char_layer_name == "none": # the character layer to make visible on completion
        active_layer = image.active_layer
        pdb.gimp_item_set_visible(original_active_layer, True)
        pdb.gimp_image_set_active_layer(image,original_active_layer)
		
		
    if not apply_to_custom: # the layer to make visible on completion for non-custom font mode
        pdb.gimp_layer_set_offsets(original_active_layer,0,0)
        pdb.gimp_item_set_visible(original_active_layer, True)
        active_layer_id = proc_metal(image, original_active_layer, shade_of_grey, fill_gradient, do_gradient, metal_colour, metal_texture, do_sheen, bump_depth, effect_elevation, effect_mode_ptr, retain_original, compass, light_amount, apply_lighting, artifact_radius, artifact_delta, other_layers, pitted_drawable_name, final_bump_depth, apply_to_custom, texture_opacity, metal_curves, plate_softness, plate_border)
        pdb.gimp_layer_set_offsets(active_layer_id,offset_x,offset_y)			
        
    temp = pdb.gimp_displays_flush()
		
    # Set end undo point
    pdb.gimp_image_undo_group_end(image)

    # Restore the GIMP's settings
    gimp.context_pop()
	       

######################################################################################################################################

def fn_check_version():
    check = False
    gimp_version = pdb.gimp_version()
    string_loop = 0
    string_pos = gimp_version.find(".",string_loop)
    major_version = int(gimp_version[string_loop:string_pos])
    string_loop = string_pos + 1
    string_pos = gimp_version.find(".",string_loop)
    minor_version = int(gimp_version[string_loop:string_pos])
    string_loop = string_pos + 1
    string_pos = len(gimp_version)
    current_release = int(gimp_version[string_loop:string_pos])
    if (major_version >=2 and minor_version >= 10 and current_release >=18) or major_version > 2: check = True
    return check



def proc_do_bump(image, other_layers, bumpmap_layer_id, bumpmap_layer_name, active_layer, black, white, bumpmap_width, waterlevel, ambient, effect_elevation, effect_azimuth, bevel_mode_ptr):
    if other_layers > 0:
        work_layer = active_layer
        work_layer_name = work_layer.name
        width = work_layer.width
        height = work_layer.height
        bumpmap_width = bumpmap_width * 0.65
        if bumpmap_width < 1: bumpmap_width = 1
		
        if waterlevel > 0:
            waterlevel = waterlevel/100
            waterlevel = 1.0 - waterlevel
        else:
            waterlevel = 0.0
			
        if ambient > 0:
            ambient = ambient/100
            ambient = 1.0 - ambient
        else:
            ambient = 0.0        

        pdb.plug_in_bump_map(image, work_layer, bumpmap_layer_id, effect_azimuth, effect_elevation, bumpmap_width, 0, 0, waterlevel, ambient, True, False, 0)
		
        active_layer = image.active_layer 
        return active_layer
        
######################################################################################################################################

def proc_metal(image, active_layer, shade_of_grey, fill_gradient, do_gradient, custom_colour, metal_texture, do_sheen, bump_depth, effect_elevation, effect_mode_ptr, retain_original, compass, light_amount, apply_lighting, artifact_radius, artifact_delta, other_layers, pitted_drawable_name, final_bump_depth, apply_to_custom, texture_opacity, metal_curves, plate_softness, plate_border):

    effect_azimuth, light_pos_x, light_pos_y, light_dir_x, light_dir_y = fn_compass(compass, light_amount) # lighting plug-in values taken from filter dialog

    work_layer = active_layer # ie character layer id
    work_layer_name = work_layer.name # ie character layer name
    width = work_layer.width
    height = work_layer.height
			
    current_fg = pdb.gimp_context_get_foreground() # make a note of the current foreground colour
	
    # set bumpmap amount from percentage in filter dialog
    bump_depth = bump_depth * 0.65
    if bump_depth < 1.0: bump_depth = 1.0
    waterlevel = 0
    ambient = 0.5
    if final_bump_depth>0:
        final_bump_depth = final_bump_depth * 0.65
        if final_bump_depth < 1.0: final_bump_depth = 1.0
		
		
    if metal_texture == 6: #### Plate
        size_factor = height/60 # for shrinking outline
        #size_factor = height/(plate_border*15) # for shrinking outline
        size_factor = height/(plate_border*12) # for shrinking outline
        #size_factor = (height/3)/(5*(10 - plate_softness)) # for shrinking outline
			
        control_layer = pdb.gimp_layer_copy(work_layer, True)
        image.add_layer(control_layer, -1)
        pdb.gimp_item_set_visible(control_layer, False)
        pdb.gimp_item_set_name(control_layer,"Control Layer")
		
		
    if not retain_original:
			
        temp_layer = gimp.Layer(image, "Temp Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the darker layer for white bases
        image.add_layer(temp_layer, -1)
        metal_grey = fn_get_bg_bump_col(shade_of_grey)
        pdb.gimp_context_set_foreground(metal_grey)
		
        if metal_texture == 6: pdb.gimp_context_set_foreground(white)####
        pdb.gimp_image_select_item(image, 2, work_layer) # get character outline ####
        if metal_texture == 6: pdb.gimp_selection_shrink(image, size_factor) ####
		
        pdb.gimp_drawable_edit_fill(temp_layer, 0) # fill character outline
        pdb.gimp_selection_none(image)
        pdb.gimp_drawable_edit_clear(work_layer)
		
        work_layer = pdb.gimp_image_merge_visible_layers(image,0)
    
	
	
    pdb.gimp_context_set_foreground(black) 
    metal_layer = gimp.Layer(image, "Metal Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the metallic layer
    image.add_layer(metal_layer, -1)
    pdb.gimp_drawable_fill(metal_layer, 0) # fill entire layer
	
    pdb.gimp_image_select_item(image, 2, work_layer)
    if metal_texture == 6 and retain_original: pdb.gimp_selection_shrink(image, size_factor) ####
	
    # get y co-ordinate values for blending
    current_selection, cs_x1, cs_y1, cs_x2, cs_y2, cs_width, cs_height  = fn_selection_info(image)
    blend_x0 = cs_x1
    blend_x1 = cs_x2
    blend_y0 = cs_y1
    blend_y1 = cs_y2
    sel_blur_factor = cs_height * 0.05 # base initial blur on height of actual character within layer
	
    pdb.gimp_context_set_foreground(white)
    pdb.gimp_drawable_edit_fill(metal_layer, 0) # fill character outline
	
    pdb.gimp_selection_none(image)
	
    # make a copy of the original character so that we can use it later to meld with the newly applied lighting effects
    copy_of_metal = pdb.gimp_layer_copy(metal_layer, True)
    image.add_layer(copy_of_metal, -1)
    pdb.gimp_item_set_visible(copy_of_metal, False)
    pdb.gimp_item_set_name(copy_of_metal,"Copy of Metal")
	
    if metal_texture == 6: # Plate
        #pdb.plug_in_gauss(image, copy_of_metal, float(height/50), float(height/50), 1) # blur linked to height of custom font
        pdb.plug_in_gauss(image, copy_of_metal, (cs_height * (plate_softness/100) ), (cs_height * (plate_softness/100) ), 1)
	
    else:
        pdb.plug_in_gauss(image, copy_of_metal, (sel_blur_factor), (sel_blur_factor), 1)
        #pdb.plug_in_gauss(image, copy_of_metal, float(height/50), float(height/50), 1) # blur linked to height of custom font
		
    final_copy = pdb.gimp_layer_copy(copy_of_metal, True)
    image.add_layer(final_copy, -1)
    pdb.gimp_item_set_name(final_copy,"Final Copy")
	

    if metal_texture <> 8: # dummy value
        # create a temporary layer to get co-ordinates for gradients
        copy_layer = pdb.gimp_layer_copy(metal_layer, TRUE)
        temp = pdb.gimp_image_insert_layer(image, copy_layer, None, -1)
        pdb.plug_in_autocrop_layer(image, copy_layer)
        pdb.gimp_drawable_edit_clear(copy_layer)
        copy_layer.fill(1)
        pdb.gimp_image_select_item(image, 2, copy_layer)
        current_selection, cs_x1, cs_y1, cs_x2, cs_y2  = pdb.gimp_selection_bounds(image)
        blend_x0, blend_y0, blend_x1, blend_y1 = fn_gradient_direction(compass, width, cs_x1, cs_y1, cs_x2, cs_y2)
        pdb.gimp_selection_none(image)
        image.remove_layer(copy_layer)
		
		
    pdb.gimp_drawable_edit_clear(metal_layer)
	
    ######## Texture specifics go here #######
	
    if metal_texture == 0: #Glitter texture
		
        if retain_original:
	
            # make a copy of the original character and desaturate it ready for bumpmapping
            position = pdb.gimp_image_get_item_position(image, work_layer)
            temp_copy = pdb.gimp_layer_copy(work_layer, True)
            image.add_layer(temp_copy, position)
            pdb.gimp_item_set_visible(temp_copy, True)
            pdb.gimp_drawable_desaturate(temp_copy,0)		
        
        # fill the character outline with colour ready for bumpmapping
        pdb.gimp_image_select_item(image, 2, work_layer)		
		
        bright_bg_grey, bright_fg_grey = fn_set_fg_bg_blend_colours(shade_of_grey)
        
        pdb.gimp_context_set_background(bright_bg_grey)
        pdb.gimp_context_set_foreground(bright_fg_grey)			
			
        pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
        random.seed()
        pdb.gimp_selection_none(image)
        pdb.plug_in_hsv_noise(image,metal_layer, 1,180,255,255)
        if not retain_original:pdb.plug_in_hsv_noise(image,metal_layer, 1,180,255,255)
        pdb.gimp_drawable_desaturate(metal_layer,0)
			
			
        if retain_original:
	
            pdb.gimp_layer_set_opacity(metal_layer, texture_opacity)
            mode = 45 # soft light
            metal_layer.mode = mode
            metal_layer = pdb.gimp_image_merge_down(image, metal_layer, 0)
		
			
    if metal_texture == 1 or metal_texture == 6: # Bright/Plate Texture
	
        if retain_original:
            # make a copy of the original character and desaturate it ready for bumpmapping
            copy = pdb.gimp_edit_copy(work_layer)
            floater = pdb.gimp_edit_paste(metal_layer ,True)
            pdb.gimp_floating_sel_anchor(floater)
            pdb.gimp_drawable_desaturate(metal_layer,0)
			
			
            temp_copy = pdb.gimp_layer_copy(metal_layer, True)
            image.add_layer(temp_copy, -1)
            pdb.gimp_item_set_name(temp_copy,"Temp Copy")
            pdb.gimp_item_set_visible(temp_copy, True)
            random.seed()
            amount = random.randrange(100,10000000,1)
			
            if metal_texture<>6: pdb.gimp_layer_set_opacity(temp_copy, texture_opacity)
            
            if metal_texture == 6: ####
			
                mode = 28 #### Normal
            else:
                mode = 45 # soft light
				
            temp_copy.mode = mode
            #metal_layer = pdb.gimp_image_merge_down(image, temp_copy, 0)
		
        else:
            pdb.gimp_image_select_item(image, 2, work_layer)
		
            pdb.gimp_drawable_edit_fill(metal_layer, 0) # fill character outline
			
        copy_layer = pdb.gimp_layer_copy(metal_layer, TRUE)
        temp = pdb.gimp_image_insert_layer(image, copy_layer, None, -1)
        pdb.gimp_selection_none(image)
        pdb.gimp_image_set_active_layer(image, copy_layer)
        pdb.plug_in_autocrop_layer(image, copy_layer)
        pdb.gimp_drawable_edit_fill(copy_layer, 0) # fill character outline
        pdb.gimp_image_select_item(image, 2, copy_layer)
        # get y co-ordinate values for blending
        current_selection, cs_x1, cs_y1, cs_x2, cs_y2  = pdb.gimp_selection_bounds(image)
        blend_x0, blend_y0, blend_x1, blend_y1 = fn_gradient_direction(compass, width, cs_x1, cs_y1, cs_x2, cs_y2)
        pdb.gimp_selection_none(image)
        image.remove_layer(copy_layer)
        gradient_mode = fn_set_bevel_mode(effect_mode_ptr)		
		
        if retain_original:
            temp_layer = gimp.Layer(image, "Temp Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the darker layer for white bases
            image.add_layer(temp_layer, -1)
        else:
            pdb.gimp_drawable_edit_clear(metal_layer)
	
            if metal_texture<> 6:pdb.gimp_image_select_item(image, 2, work_layer)
            #pdb.gimp_context_set_foreground(lightgrey)
            pdb.gimp_context_set_foreground(verylightgrey)
        
        bright_bg_grey, bright_fg_grey = fn_set_fg_bg_blend_colours(shade_of_grey)
		
        pdb.gimp_context_set_background(bright_bg_grey)
        pdb.gimp_context_set_foreground(bright_fg_grey)
        pdb.gimp_context_set_gradient_fg_bg_rgb()
		
        if retain_original:
            temp_layer.mode = gradient_mode
            pdb.gimp_drawable_edit_gradient_fill(temp_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
            if metal_texture<> 6: metal_layer = pdb.gimp_image_merge_down(image, temp_layer, 0) ####
        else:
            pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
			
        pdb.gimp_selection_none(image)
		
    if metal_texture == 2: # Verdigris Texture	
	
        if retain_original:
	
            # make a copy of the original character and desaturate it ready for bumpmapping
            position = pdb.gimp_image_get_item_position(image, work_layer)
            temp_copy = pdb.gimp_layer_copy(work_layer, True)
            image.add_layer(temp_copy, position)
            pdb.gimp_item_set_visible(temp_copy, True)
            pdb.gimp_drawable_desaturate(temp_copy,0)
	
	
        pdb.gimp_image_select_item(image, 2, work_layer)
				
        bright_bg_grey, bright_fg_grey = fn_set_fg_bg_blend_colours(shade_of_grey)
        
        pdb.gimp_context_set_background(bright_bg_grey)
        pdb.gimp_context_set_foreground(bright_fg_grey)			
			
        pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)

        pdb.gimp_selection_none(image)
		
        pdb.plug_in_plasma(image,metal_layer, random.randrange(1000000, 2000000), 7)
        pdb.gimp_image_select_item(image, 2, work_layer)
        temp_width = image.width
        temp_height = image.height
        temp_selection, temp_x1, temp_x2, temp_y1, temp_y2, temp_s_width, temp_s_height = fn_selection_info(image)
        pdb.gimp_selection_invert(image)
        if temp_width>temp_s_width or temp_height>temp_s_height:successful = pdb.gimp_edit_cut(metal_layer)
        pdb.gimp_selection_none(image)

        pdb.gimp_drawable_desaturate(metal_layer,0)
		
        if retain_original:
            pdb.gimp_layer_set_opacity(metal_layer, texture_opacity)
            mode = 50 # linear light
            temp_copy.mode = mode
            metal_layer = pdb.gimp_image_merge_down(image, metal_layer, 0)
		
        active_layer = image.active_layer 
        #pdb.plug_in_sel_gauss(image, active_layer, artifact_radius, artifact_delta)
        metal_colour = active_layer
		
			
			
    if metal_texture == 3: # Dull Texture
		
        if retain_original:
            copy = pdb.gimp_edit_copy(work_layer)
            floater = pdb.gimp_edit_paste(metal_layer ,True)
            pdb.gimp_floating_sel_anchor(floater)
            pdb.gimp_drawable_desaturate(metal_layer,0)			
			
            temp_copy = pdb.gimp_layer_copy(metal_layer, True)
            image.add_layer(temp_copy, -1)
            pdb.gimp_item_set_name(temp_copy,"Temp Copy")
            pdb.gimp_item_set_visible(temp_copy, True)
            random.seed()
            amount = random.randrange(100,10000000,1)			
			
            pdb.gimp_image_select_item(image, 2, work_layer)
            pdb.gimp_context_set_foreground(lightgrey)
            pdb.gimp_context_set_background(verydarkgrey)
            pdb.gimp_context_set_gradient_fg_bg_rgb()
            pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
            random.seed()
		
            pdb.gimp_selection_none(image)
			
			
            pdb.gimp_layer_set_opacity(temp_copy, 100 - texture_opacity)
            mode = 45 # soft light
            temp_copy.mode = mode
            metal_layer = pdb.gimp_image_merge_down(image, temp_copy, 0)
			
			
        else:
            # fill the character outline with colour ready for bumpmapping
            pdb.gimp_image_select_item(image, 2, work_layer)
			
            bright_bg_grey, bright_fg_grey = fn_set_fg_bg_blend_colours(shade_of_grey)
        
            pdb.gimp_context_set_background(bright_bg_grey)
            pdb.gimp_context_set_foreground(bright_fg_grey)
			
            pdb.gimp_context_set_gradient_fg_bg_rgb()
            pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
            random.seed()
		
            pdb.gimp_selection_none(image)	
			
    if metal_texture == 4: # Annealed Texture	
		
        if retain_original:
	
            # make a copy of the original character and desaturate it ready for bumpmapping
            position = pdb.gimp_image_get_item_position(image, work_layer)
            temp_copy = pdb.gimp_layer_copy(work_layer, True)
            image.add_layer(temp_copy, position)
            pdb.gimp_item_set_visible(temp_copy, True)
            pdb.gimp_drawable_desaturate(temp_copy,0)
			
        pdb.gimp_image_select_item(image, 2, work_layer)
				
        bright_bg_grey, bright_fg_grey = fn_set_fg_bg_blend_colours(shade_of_grey)
        
        pdb.gimp_context_set_background(bright_bg_grey)
        pdb.gimp_context_set_foreground(bright_fg_grey)			
			
        pdb.gimp_drawable_edit_gradient_fill(metal_layer, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)

        random.seed()
        amount = random.randrange(100,10000000,1)
        pdb.plug_in_solid_noise(image,metal_layer, False, True, amount, 15,1,1)
        pdb.gimp_drawable_desaturate(metal_layer,0)
        pdb.gimp_selection_none(image)
		
		
        if retain_original:
            pdb.gimp_layer_set_opacity(metal_layer, texture_opacity)
            mode = 45 # soft light
            metal_layer = pdb.gimp_image_merge_down(image, metal_layer, 0)
	
    if metal_texture == 5: # Brushed Texture
	
        if retain_original:
	
            # make a copy of the original character and desaturate it ready for bumpmapping
            copy = pdb.gimp_edit_copy(work_layer)
            floater = pdb.gimp_edit_paste(metal_layer ,True)
            pdb.gimp_floating_sel_anchor(floater)
            pdb.gimp_drawable_desaturate(metal_layer,0)			
			
            temp_copy = pdb.gimp_layer_copy(metal_layer, True)
            image.add_layer(temp_copy, -1)
            pdb.gimp_item_set_name(temp_copy,"Temp Copy")
            pdb.gimp_item_set_visible(temp_copy, True)
            random.seed()
            amount = random.randrange(4,20,1)
			
            pdb.gimp_image_select_item(image, 2, work_layer)
            pdb.plug_in_hsv_noise(image,temp_copy, 1,180,255,255)
            pdb.plug_in_hsv_noise(image,temp_copy, 1,180,255,255)
            pdb.plug_in_wind(image,temp_copy, 10,0,amount,1,1)
            pdb.plug_in_wind(image,temp_copy, 10,0,amount,1,1)
            pdb.gimp_selection_none(image)
            pdb.gimp_drawable_desaturate(temp_copy,0)
			
            pdb.gimp_layer_set_opacity(temp_copy, texture_opacity)
            mode = 23 # overlay
            temp_copy.mode = mode
            metal_layer = pdb.gimp_image_merge_down(image, temp_copy, 0)
		
        else:
        
            # fill the character outline with colour ready for bumpmapping
            #pdb.gimp_image_select_item(image, 2, work_layer)
            pdb.gimp_context_set_foreground(verylightgrey)
            pdb.gimp_drawable_edit_fill(metal_layer, 0) # fill character outline with grey
            pdb.plug_in_hsv_noise(image,metal_layer, 1,180,255,255)
            pdb.plug_in_hsv_noise(image,metal_layer, 1,180,255,255)
            random.seed()
            amount = random.randrange(4,20,1)
            pdb.plug_in_wind(image,metal_layer, 10,0,amount,0,1)
            pdb.plug_in_wind(image,metal_layer, 10,0,amount,0,1)
            #pdb.gimp_selection_none(image)
            pdb.gimp_drawable_desaturate(metal_layer,0)
            pdb.gimp_image_select_item(image, 2, work_layer)
            pdb.gimp_selection_invert(image)
            success=pdb.gimp_edit_cut(metal_layer)
            pdb.gimp_selection_none(image)
	
    ######## Back to global texture code ########
	
    if metal_texture == 6: # Plate
        if retain_original:
            pdb.plug_in_bump_map(image, temp_layer, copy_of_metal, effect_azimuth, effect_elevation, bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
            chrome_mask = pdb.gimp_layer_create_mask(temp_layer, 1) # create mask filled with black for metal layer
            pdb.gimp_layer_add_mask(temp_layer, chrome_mask) # add mask to layer
            pdb.gimp_image_select_item(image, 2, control_layer)
            pdb.gimp_context_set_background(white)
            pdb.gimp_drawable_edit_fill(chrome_mask, 1) # fill character outline with white
            pdb.gimp_selection_none(image)
            temp_layer.mode = 59 # merge mode ####
            pdb.gimp_item_set_visible(temp_copy, False) ####
            metal_layer = pdb.gimp_image_merge_down(image, temp_layer, 0)

        else:
            pdb.plug_in_bump_map(image, metal_layer, copy_of_metal, effect_azimuth, effect_elevation, bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
		
            chrome_mask = pdb.gimp_layer_create_mask(metal_layer, 1) # create mask filled with black for metal layer
            pdb.gimp_layer_add_mask(metal_layer, chrome_mask) # add mask to layer
            pdb.gimp_image_select_item(image, 2, control_layer)
            pdb.gimp_context_set_background(white)
            pdb.gimp_drawable_edit_fill(chrome_mask, 1) # fill character outline with white
            pdb.gimp_selection_none(image)
		
        image.remove_layer(control_layer)

    else:
        pdb.plug_in_bump_map(image, metal_layer, copy_of_metal, effect_azimuth, effect_elevation, bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
		
    if metal_texture <> 5: # if not brushed
        if artifact_radius>0.0: pdb.plug_in_sel_gauss(image, metal_layer, artifact_radius, artifact_delta) # Trying this blur here but not 
    #pdb.plug_in_sel_gauss(image, metal_layer, artifact_radius, artifact_delta) # Trying this blur here but not 
    #pdb.plug_in_sel_gauss(image, metal_layer, 5.00, 0.200)
	
    active_layer = image.active_layer 
	
    if metal_texture == 5: # brushed again
        if final_bump_depth == 0: 
            if artifact_radius>0.0: pdb.plug_in_sel_gauss(image, active_layer, artifact_radius, artifact_delta)
		
		
    metal_colour = active_layer
	
	
    # REM: Set lighting effect to metal_layer using copy_of_metal for bump map
    # Set variables for proc_lighting
    bump_map = metal_layer
    #if retain_original: bump_map = metal_layer
    env_map = None
    bump_map_flag = True
    env_map_flag = False
    bump_map_curve = 0
    light_source = 1
    light_colour = (255,255,255)
    #light_pos_x = -1.00
    #light_pos_y = -1.00
    light_pos_z = 1.00
    #light_dir_x = -1.50
    #light_dir_y = -1.50
    light_dir_z = 1.00
    light_glow = 0.80 # 0.60
    light_bright =  0.50 # 0.50
    light_diffuse_reflec = 3.00 #2.80
    light_shiny = 0.20 # 1.00 0.10
    light_polished = 30 #27 20
    anti_aliasing = True
    new_light_image = False
    transparent_bg = False
	
    if apply_lighting:
	    #REM: Call Lighting plugin
        proc_lighting(image, metal_layer, bump_map, env_map, bump_map_flag, env_map_flag, bump_map_curve, light_source, light_colour, 
        light_pos_x, light_pos_y, light_pos_z,
        light_dir_x, light_dir_y, light_dir_z, 
        light_glow, light_bright, light_diffuse_reflec, light_shiny, light_polished, 
        anti_aliasing, new_light_image, transparent_bg)
	
    temp = pdb.gimp_displays_flush()
		
    random.seed()
	
    ######## Texture specific curves values
	
    #Rem Prepare texture specific curves values
	
    if metal_texture == 0: #Glitter curves
        curves_list = [0.0,0.0,0.1725,0.3215,0.3058,0.1215,0.4549,0.8196,0.5529,0.298,0.647,1.0]
    
        random_curves_list = [0.0,0.0,0.1725,0.3215,0.3058,0.1215,0.4549,float(random.randrange(820,996,1)/1000),0.5529,float(random.randrange(27,298,1)/1000),0.647,1.0]

    elif metal_texture == 1: # Bright curves
        curves_list = [0.0,0.0,0.0352,0.345,0.2588,1.0,0.4117,0.2235,0.5058,0.996,0.5882,0.2862,0.7137,0.9921,0.7607,0.549,0.8313,0.9255,0.9137,0.745,1.0,1.0]
        random_curves_list = [0.0,0.0,0.0352,0.345,0.2588,float(random.randrange(823,1000,1)/1000),float(random.randrange(333,470,1)/1000),0.2235,0.5058,0.996,0.5882,0.2862,0.7137,float(random.randrange(784,992,1)/1000),0.7607,0.549,0.8313,0.9255,0.9137,0.745,1.0,1.0]
			
    elif metal_texture == 2: # Verdigris curves
        curves_list = [0.0,0.0,0.0705,0.0,0.631,1.0,0.729,0.0,0.8313,1.0,0.9215,0.0,1.0,1.0] 
        random_curves_list = [0.0,0.0,0.541,0.0,0.631,float(random.randrange(1,1000,1)/1000),0.729,0.0,0.831,1.0,0.921,0.0,1.0,1.0]
    
    elif metal_texture == 3: # Dull curves
        curves_list = [0.0, 0.0, 0.133,0.3176,0.2235,0.0823,0.3255,0.466,0.4039,0.2235,0.4705,0.7098,0.592,0.4078,0.6117,0.8745,0.8,0.2627,1.0, 1.0]
        random_curves_list = [0.0, 0.0, 0.133,0.3176,0.0823,0.3255,0.466,0.4039,float(random.randrange(345,451,1)/1000),0.4705,float(random.randrange(431,549,1)/1000),0.7098,0.592,0.4078,0.6117,0.8745,0.8,0.2627,1.0, 1.0]
			
    elif metal_texture == 4: # Annealed curves
        curves_list = [0.0,0.0,0.1725,0.3215,0.3058,0.1215,0.4549,0.8196,0.5529,0.298,0.647,1.0]
        random_curves_list = [0.0,0.0,0.1725,0.3215,0.3058,0.1215,0.4549,float(random.randrange(820,996,1)/1000),0.5529,float(random.randrange(27,298,1)/1000),0.647,1.0]
			
    elif metal_texture == 5: # Brushed Texture has no curves
        curves_list = [0.0,0.0,1.0,1.0]
        random_curves_list = [0.0,0.0,1.0,1.0]

    else:
	
        if metal_curves == 0: #Glitter curves
            curves_list = [0.0,0.0,44.0,82.0,78.0,31.0,116.0,209.0,141.0,76.0,165.0,255.0]
            random_curves_list = [0.0,0.0,44.0,82.0,78.0,31.0,116.0,random.randrange(209.0,254.0,1.0),141.0,random.randrange(7.0,76.0,1.0),165.0,255.0]
			
        elif metal_curves == 1: # Bright curves
            curves_list = [0,0,9,88,66,255,105,57,129,254,150,73,182,253,194,140,212,236,233,190,255,255]
            random_curves_list = [0,0,9,88,66,random.randrange(210,255,1),random.randrange(85,120,1),57,129,254,150,73,182,random.randrange(200,253,1),194,140,212,236,233,190,255,255]
			
        elif metal_curves == 2: # Verdigris curves
            curves_list = [0,0,138,0,161,255,186,0,212,255,235,0,255,255] 
            random_curves_list = [0,0,138,0,161,random.randrange(0,255,1),186,0,212,255,235,0,255,255]
			
        elif metal_curves == 3: # Dull curves
            curves_list = [0, 0, 34,81,57,21,83,119,103,57,120,181,151,104,156,223,204,67,255, 255]
            random_curves_list = [0, 0, 34,81,57,21,83,119,random.randrange(88,115,1),57,random.randrange(110,140,1),181,151,104,156,223,204,67,255, 255]
			
        elif metal_curves == 4: # Annealed curves
            curves_list = [0,0,44,82,78,31,116,209,141,76,165,255]
            random_curves_list = [0,0,44,82,78,31,116,random.randrange(209,254,1),141,random.randrange(7,76,1),165,255]
			
        elif metal_curves == 5: # Plated - Polish 1
            curves_list = [0.0,0.0,0.247,0.996,0.486,0.231,0.666,1.0,0.8627,0.49,1.0,1.0]
            random_curves_list = [0.0,0.0,0.247,0.996,0.486,0.231,0.666,1.0,0.8627,0.49,1.0,1.0] # Random same as normal
			
        elif metal_curves == 6: # Plated - Polish 2
            curves_list = [0.0,0.0,0.18,0.9725,0.345,0.0,0.6,0.0196,0.7568,1.0,0.8549,0.5529,1.0,1.0]
            random_curves_list = [0.0,0.0,0.18,0.9725,0.345,0.0,0.6,0.0196,0.7568,1.0,0.8549,0.5529,1.0,1.0]
		
        elif metal_curves == 7: # Plated - Polish 3
            curves_list = [0.0,0.0,0.1176,0.5019,0.196,0.228,0.3058,0.792,0.4078,0.58,0.537,0.847,0.65,0.243,0.74,0.69,0.8588,0.345,1.0,1.0]
            random_curves_list = [0.0,0.0,0.1176,0.5019,0.196,0.228,0.3058,0.792,0.4078,0.58,0.537,0.847,0.65,0.243,0.74,0.69,0.8588,0.345,1.0,1.0]
		
        elif metal_curves == 8: # Plated - Polish 4
            curves_list = [0.0,0.0,0.2117,0.9607,0.4275,0.4549,0.5765,0.8901,0.6784,0.5725,0.8156,0.949,0.93,0.7568,1.0,1.0]
            random_curves_list = [0.0,0.0,0.2117,0.9607,0.4275,0.4549,0.5765,0.8901,0.6784,0.5725,0.8156,0.949,0.93,0.7568,1.0,1.0]
		
        elif metal_curves == 9: # Plated -  Polish 5
            curves_list = [0.0,0.0,0.486,0.0,0.6549,1.0,0.694,0.0,0.8039,0.99,0.9215,0.0,0.9725,0.988,1.0,1.0]
            random_curves_list = [0.0,0.0,0.486,0.0,0.6549,1.0,0.694,0.0,0.8039,0.99,0.9215,0.0,0.9725,0.988,1.0,1.0]
			
        elif metal_curves == 10: # Plated -  Polish 6
            curves_list = [0.0,0.0,0.3765,0.247,0.5215,0.847,0.7058,0.44,1.0,1.0]
            random_curves_list = [0.0,0.0,0.3765,0.247,0.5215,0.847,0.7058,0.44,1.0,1.0]
			
        elif metal_curves == 11: # Plated - Polish 7
            curves_list = [0.0,0.0,0.24,0.137,0.36,0.526,0.498,0.502,0.502,0.3765,0.741,0.5176,0.86,1.0,0.878,0.506,1.0,1.0]
            random_curves_list = [0.0,0.0,0.24,0.137,0.36,0.526,0.498,0.502,0.502,0.3765,0.741,0.5176,0.86,1.0,0.878,0.506,1.0,1.0]
			
        elif metal_curves == 12: # Polish 8
            curves_list = [0.0,0.0,0.3568,0.243,0.67,0.996,0.85,0.529,1.0,1.0]
            random_curves_list = [0.0,0.0,0.3568,0.243,0.67,0.996,0.85,0.529,1.0,1.0]
	


    ######## Back to mainly global texture code ########
	
    #Rem Add curves    
    if do_sheen:
        pdb.gimp_drawable_curves_spline(metal_layer, 0, len(random_curves_list), random_curves_list)
    else:            
        pdb.gimp_drawable_curves_spline(metal_layer, 0, len(curves_list), curves_list)		
	
    pdb.gimp_drawable_edit_clear(copy_of_metal)	
	
    if metal_texture == 2: # Verdigris Texture
        if other_layers>0 and not do_gradient:
            old_theshold = pdb.gimp_context_get_sample_threshold_int()
            new_threshold = random.randrange(80,121,1)
            pdb.gimp_context_set_sample_threshold_int(new_threshold)
            pdb.gimp_image_select_color(image, 2, metal_layer, black)
            verdigris_selection = pdb.gimp_selection_save(image)
            pdb.gimp_item_set_name(verdigris_selection,"Verdigris Selection")
            pdb.gimp_selection_none(image)	
            pdb.gimp_context_set_sample_threshold_int(old_theshold)


    gradient_mode = fn_set_bevel_mode(effect_mode_ptr)
	
    # fill the selection with the gradient
    if do_gradient:
        pdb.gimp_context_set_gradient(fill_gradient)
        pdb.gimp_drawable_edit_gradient_fill(copy_of_metal, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)
    else:
	    # REM Prepare colours for custom gradient
        #Add colour and layer mode information
        gradient_mode, name_colour = fn_set_metal_colour(custom_colour)
        pdb.gimp_context_set_gradient_fg_bg_rgb()				
        # Do blending
        pdb.gimp_drawable_edit_gradient_fill(copy_of_metal, 0, 0, False, 1, 0, True, blend_x0, blend_y0, blend_x1, blend_y1)	
	
	
    pdb.gimp_item_set_visible(copy_of_metal, True)
    copy_of_metal.mode = gradient_mode
    pdb.gimp_drawable_edit_clear(work_layer)
    work_layer = pdb.gimp_image_merge_visible_layers(image,0)
	
    if final_bump_depth>0: pdb.plug_in_bump_map(image, work_layer, final_copy, effect_azimuth, effect_elevation, final_bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
    image.remove_layer(final_copy)
		
    if metal_texture == 2: # Verdigris Texture
        if other_layers>0 and not do_gradient:
            pdb.gimp_image_select_item(image, 2, verdigris_selection)
            pitted_drawable = pdb.gimp_image_get_layer_by_name(image, pitted_drawable_name)
            success = pdb.gimp_edit_copy(pitted_drawable)
            floater = pdb.gimp_edit_paste(work_layer, False)
            pdb.gimp_floating_sel_anchor(floater)
            image.remove_channel(verdigris_selection)
            layer_id = pdb.gimp_image_get_active_layer(image)
            active_layer = image.active_layer 
			
    if metal_texture == 6: # Plate
        if retain_original:
            #temp_copy.mode = 23 # overlay mode ####
            pdb.gimp_item_set_visible(temp_copy, True) ####
            pdb.gimp_image_lower_item(image, temp_copy)
            pdb.gimp_layer_set_opacity(work_layer, texture_opacity) ####
            work_layer = pdb.gimp_image_merge_visible_layers(image,0) ####
            pdb.gimp_item_set_name(work_layer, work_layer_name)
			
    pdb.gimp_context_set_foreground(current_fg) # re-set foreground colour
	
    active_layer = image.active_layer 
    return active_layer	

        
######################################################################################################################################
  # -------------------------------------------------------------------------------------
	
def proc_lighting(light_image, light_drawable, b_map, e_map, b_map_flag, e_map_flag, b_map_curve, l_source, l_colour, l_pos_x, l_pos_y, l_pos_z, l_dir_x, l_dir_y, l_dir_z, l_glow, l_bright, l_diffuse_reflec, l_shiny, l_polished, a_aliasing, n_image, mk_transp_bg):
    
    # Rem: Use the lighting Plugin
    #LIGHT 1 

    pdb.plug_in_lighting(light_image, light_drawable, b_map, e_map,  b_map_flag, e_map_flag, b_map_curve, l_source, l_colour, l_pos_x, l_pos_y, l_pos_z, l_dir_x, l_dir_y, l_dir_z, l_glow, l_bright, l_diffuse_reflec, l_shiny, l_polished, True, False, False)

    # -------------------------------------------------------------------------------------


        
######################################################################################################################################

def fn_set_fg_bg_blend_colours(shade_of_grey):

    shade_of_grey = abs(shade_of_grey - 100)
    if shade_of_grey<80:
        bg_grey = shade_of_grey
        fg_grey = 90
    else:
        bg_grey = shade_of_grey
        fg_grey = 100
    bg_grey_colour = int(bg_grey * 2.55)
    fg_grey_colour = int(fg_grey * 2.55)
        
    bright_bg_grey = gimpcolor.RGB(bg_grey_colour,bg_grey_colour,bg_grey_colour)
    bright_fg_grey = gimpcolor.RGB(fg_grey_colour,fg_grey_colour,fg_grey_colour)

    return bright_bg_grey, bright_fg_grey
        
######################################################################################################################################


def fn_gradient_direction(compass, width, cs_x1, cs_y1, cs_x2, cs_y2):

 
    if compass == 0:
        # North
        blend_x0 = width/2
        blend_y0 = cs_y1
        blend_x1 = width/2
        blend_y1 = cs_y2
    elif compass == 1:
        # North East
        blend_x0 = width
        blend_y0 = cs_y1
        blend_x1 = 0
        blend_y1 = cs_y2
    elif compass == 2:
        # East
        blend_x0 = width
        blend_y0 = cs_y1 + ((cs_y2-cs_y1)/2)
        blend_x1 = 0
        blend_y1 = cs_y1 + ((cs_y2-cs_y1)/2)
    elif compass == 3:
        # South East
        blend_x0 = width
        blend_y0 = cs_y2
        blend_x1 = 0
        blend_y1 = cs_y1
    elif compass == 4:
        # South
        blend_x0 = width/2
        blend_y0 = cs_y2
        blend_x1 = width/2
        blend_y1 = cs_y1
    elif compass == 5:
        # South West
        blend_x0 = 0
        blend_y0 = cs_y2
        blend_x1 = width
        blend_y1 = cs_y1
    elif compass == 6:
        # West
        blend_x0 = 0
        blend_y0 = ((cs_y2-cs_y1)/2)
        blend_x1 = width
        blend_y1 = ((cs_y2-cs_y1)/2)
    else:
        # North West
        blend_x0 = cs_x1
        blend_y0 = cs_y1
        blend_x1 = cs_x2
        blend_y1 = cs_y2
		
    return blend_x0, blend_y0, blend_x1, blend_y1
       
######################################################################################################################################


def fn_compass(compass, light_amount):

    # Set lighting plug-in pos and dir xyz values from compass point selected via dialog UI
    if compass == 0:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 0,0,-1,-1 #S
        effect_azimuth = 90
        compass_info = "N"
    elif compass == 1:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,-1,-1 #SW
        effect_azimuth = 45
        compass_info = "NE"
    elif compass == 2:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,0,0 #W
        effect_azimuth = 0
        compass_info = "E"
    elif compass == 3:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 1,1,1,1 #NW
        effect_azimuth = 315
        compass_info = "SE"
    elif compass == 4:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = 0,0,1,1 #N
        effect_azimuth = 270
        compass_info = "S"
    elif compass == 5:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,1,1 #NE
        effect_azimuth = 225
        compass_info = "SW"
    elif compass == 6:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,0,0 #E
        effect_azimuth = 180
        compass_info = "W"
    else:
        light_pos_x,light_dir_x,light_pos_y,light_dir_y = -1,-1,-1,-1 #SE
        effect_azimuth = 135
        compass_info = "NW"
		
    # set non UI bumpmap plugin parameters here to reduce height of dialog
    effect_elevation = 35 # possible temp elevation
    #bumpmap_depth = 10 # possible temp bumpmap depth
    waterlevel = 0
    ambient = 0
    
    light_pos_x *= light_amount
    light_pos_y *= light_amount
    light_dir_x *= light_amount
    light_dir_y *= light_amount
	
	
    #light_pos_z, light_dir_z = light_z,light_z
	
    return effect_azimuth, light_pos_x, light_pos_y, light_dir_x, light_dir_y
	
######################################################################################################################################	


def fn_get_bg_bump_col(grey):
    if grey == 0:
        grey_shade = gimpcolor.RGB(222,222,222)
        grey_info = "Lightest grey"
    elif grey == 1:
        grey_shade = gimpcolor.RGB(170, 170, 170)
        grey_info = "Very light grey"
    elif grey == 2:
        grey_shade = gimpcolor.RGB(125, 125, 125)
        grey_info = "Light grey"
    elif grey == 3:
        grey_shade = gimpcolor.RGB(105, 105, 105)
        grey_info = "Mid-grey"
    elif grey == 4:
        grey_shade = gimpcolor.RGB(75, 75, 75)
        grey_info = "Dark grey"
    elif grey == 5:
        grey_shade = gimpcolor.RGB(50, 50, 50)
        grey_info = "Very dark grey"
    else:
        grey_shade = gimpcolor.RGB(20, 20, 20)
        grey_info = "Darkest grey"

    return grey_shade
	
######################################################################################################################################

def fn_set_metal_colour(mc):
    if (mc == 1):
        n_colour = "Gold"
        pdb.gimp_context_set_foreground(lightgold)
        pdb.gimp_context_set_background(darkgold)
        c_mode = 19#SOFTLIGHT
    elif (mc == 2):
        n_colour = "Copper"
        pdb.gimp_context_set_foreground(darkcopper)
        pdb.gimp_context_set_background(lightcopper)
        c_mode = 21#GRAIN MERGE
    elif (mc == 3):
        n_colour = "Brass"
        pdb.gimp_context_set_foreground(darkbrass)
        pdb.gimp_context_set_background(lightbrass)
        c_mode = 19#SOFTLIGHT
    elif (mc == 4):
        n_colour = "Bronze"
        pdb.gimp_context_set_foreground(darkbronze)
        pdb.gimp_context_set_background(lightbronze)
        c_mode = 26#LCH Color new mode
        #c_mode = 45#SOFTLIGHT new mode
    else:
        n_colour = "Silver"
        pdb.gimp_context_set_foreground(lightgrey)
        #pdb.gimp_context_set_background(midgrey)
        pdb.gimp_context_set_background(darkgrey)
        c_mode = 19#SOFTLIGHT

    return c_mode, n_colour
	
######################################################################################################################################
	
def proc_emboss(image, active_layer, black, white, effect_width, effect_elevation, effect_azimuth, emboss_blur, effect_mode_ptr):
    if effect_width < 1: effect_width = 1
    grey = gimpcolor.RGB(80, 80, 80)
    work_layer = active_layer # ie character layer
    work_layer_name = work_layer.name # ie character layer id
    width = work_layer.width
    height = work_layer.height
    colour = gimp.get_background()
    emboss_mode = fn_set_bevel_mode(effect_mode_ptr)
	
    layer_blur = image.height * 0.015
    if layer_blur < 1: layer_blur = 1	
	
    pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_background(white)
	
    pdb.gimp_image_select_item(image, 2, work_layer)
	
    emboss_layer = gimp.Layer(image, "Emboss Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the bump map
    image.add_layer(emboss_layer, -1)
    emboss_layer.fill(1)
    pdb.gimp_drawable_edit_fill(emboss_layer, 0)
	
    grey_layer = gimp.Layer(image, "Grey Layer", width, height, RGBA_IMAGE, 100, 0) # layer to become the bump map
    image.add_layer(grey_layer, -1)
    pdb.gimp_context_set_foreground(midgrey)	
    #grey_layer.fill(1)
    pdb.gimp_drawable_edit_fill(grey_layer, 0)
	
    pdb.gimp_context_set_foreground(black)
    pdb.gimp_context_set_background(white)
	
    pdb.gimp_selection_none(image) ####
    pdb.plug_in_gauss_iir(image, emboss_layer, emboss_blur, True, True) # blur the bump map
	
    other_layers = 1
    bumpmap_layer_name = "Emboss Layer"
    bumpmap_layer = emboss_layer
	
    #active_layer = new_layer
	
    #temp values
    bump_waterlevel = 0
    bump_ambient = 0
    bump_depth = effect_width
    waterlevel = 0
    ambient = 0
    #end temp values
	
    if other_layers > 0:
        work_layer = active_layer
        work_layer_name = work_layer.name
        width = work_layer.width
        height = work_layer.height
        #bumpmap_width = bumpmap_width * 0.65
        #if bumpmap_width < 1: bumpmap_width = 1
		
        if waterlevel > 0:
            waterlevel = waterlevel/100
            waterlevel = 1.0 - waterlevel
        else:
            waterlevel = 0.0
			
        if ambient > 0:
            ambient = ambient/100
            ambient = 1.0 - ambient
        else:
            ambient = 0.0        

        pdb.plug_in_bump_map(image, grey_layer, emboss_layer, effect_azimuth, effect_elevation, bump_depth, 0, 0, waterlevel, ambient, True, False, 0)
    #work_layer = image.active_layer
	
    image.remove_layer(emboss_layer)
    image.lower_layer(grey_layer)
    #temp values
    blur_radius = 10
    blur_delta = 8
    #end temp values
    if artifact_radius>0.0:
        pdb.plug_in_sel_gauss(image, grey_layer, blur_radius, blur_delta)
    #pdb.plug_in_sel_gauss(image, grey_layer, 5.00, 0.200)
    work_layer.mode = emboss_mode
    pdb.gimp_image_merge_down(image, work_layer, 0)
    work_layer = image.active_layer
    pdb.gimp_item_set_name(work_layer, work_layer_name)
    #image.remove_layer(grey_layer)
		
    pdb.gimp_selection_none(image)
    active_layer = image.active_layer 
    return active_layer	
	
######################################################################################################################################
	
def fn_set_bevel_mode(bevel_mode_ptr):
    # Use New Modes when creating a bevel
    new_mode = bevel_mode_ptr
    temp_mode = 47
    if new_mode == 0:
        temp_mode = 0
        n_mode_info = "Normal"
    elif new_mode == 1:
        temp_mode = 1
        n_mode_info = "Dissolve"
    elif new_mode == 2:
        temp_mode = 57
        _mode_info = "Color Erase"
    elif new_mode == 3:
        temp_mode = 59
        n_mode_info = "Merge"
    elif new_mode == 4:
        temp_mode = 60
        n_mode_info = "Split"
    elif new_mode == 5:
        temp_mode = 36
        n_mode_info = "Lighten Only"
    elif new_mode == 6:
        temp_mode = 55
        n_mode_info = "Luma Luminance Lighten"
    elif new_mode == 7:
        temp_mode = 31
        n_mode_info = "Screen"
    elif new_mode == 8:
        temp_mode = 42
        n_mode_info = "Dodge"
    elif new_mode == 9:
        temp_mode = 33
        n_mode_info = "Addition"
    elif new_mode == 10:
        temp_mode = 35
        n_mode_info = "Darken only"
    elif new_mode == 11:
        temp_mode = 54
        n_mode_info = "Luma Luminance darken"
    elif new_mode == 12:
        temp_mode = 30
        n_mode_info = "Multiply"
    elif new_mode == 13:
        temp_mode = 43
        n_mode_info = "Burn"
    elif new_mode == 14:
        temp_mode = 53
        n_mode_info = "Linear burn"
    elif new_mode == 15:
        temp_mode = 23
        n_mode_info = "Overlay"
    elif new_mode == 16:
        temp_mode = 45
        n_mode_info = "Soft light"
    elif new_mode == 17:
        temp_mode = 44
        n_mode_info = "Hard light"
    elif new_mode == 18:
        temp_mode = 48
        n_mode_info = "Vivid light"
    elif new_mode == 19:
        temp_mode = 49
        n_mode_info = "Pin light"
    elif new_mode == 20:
        temp_mode = 50
        n_mode_info = "Linear light"
    elif new_mode == 21:
        temp_mode = 51
        n_mode_info = "Hard mix"
    elif new_mode == 22:
        temp_mode = 32
        n_mode_info = "Difference"
    elif new_mode == 23:
        temp_mode = 52
        n_mode_info = "Exclusion"
    elif new_mode == 24:
        temp_mode = 34
        n_mode_info = "Subtract"
    elif new_mode == 25:
        temp_mode = 46
        n_mode_info = "Grain extract"
    elif new_mode == 26:
        temp_mode = 47
        n_mode_info = "Grain merge"
    elif new_mode == 27:
        temp_mode = 41
        n_mode_info = "Divide"
    elif new_mode == 28:
        temp_mode = 37
        n_mode_info = "Hue HSV"
    elif new_mode == 29:
        temp_mode = 38
        n_mode_info = "Saturation HSV"
    elif new_mode == 30:
        temp_mode = 39
        n_mode_info = "Color HSL"
    elif new_mode == 31:
        temp_mode = 40
        n_mode_info = "Value HSL"
    elif new_mode == 32:
        temp_mode = 24
        n_mode_info = "Hue LCH"
    elif new_mode == 33:
        temp_mode = 25
        n_mode_info = "Chroma LCH"
    elif new_mode == 34:
        temp_mode = 26
        n_mode_info = "Color LCH"
    elif new_mode == 35:
        temp_mode = 27
        n_mode_info = "Lightness LCH"
    elif new_mode == 36:
        temp_mode = 56
        n_mode_info = "Luminance"
		
		
    bevel_mode = temp_mode
    return bevel_mode




######################################################################################################################################

def fn_test_for_custom_layer(img,l_id,l_name):
    # checks that the current layer is compatible with custom fonts
    test = False
    if len(l_name) == 1:
        test = True
    elif len(l_name) == 5 or len(l_name) == 6:
        
        if ".png" in l_name:
            test = True
            if len(l_name) == 6:
                part_name_val = unicode(l_name[0:2])
                if part_name_val <= 65535: test = False
    # select the contents of the layer
    pdb.gimp_image_select_item(img,2,l_id)
    # Test to see if the selection actually contains something
    sel_is_empty = pdb.gimp_selection_is_empty(img)
    # If there are no pixels to work with flag the layer as not viable to work on
    if sel_is_empty: test = False
    # Deselect
    pdb.gimp_selection_none(img)
    return test
	
######################################################################################################################################

######################################################################################################################################

    #### Useful Functions
# -------------------------------------------------------------------------------------
	
def fn_selection_info(im):
    # returns selection info
    selection,x1,y1,x2,y2 = pdb.gimp_selection_bounds(im)
    s_width, s_height = x2 - x1, y2 - y1
	
    return selection, x1, x2, y1, y2, s_width, s_height


######################################################################################################################################
	
def fn_selection_amount(sel_value,r_ness,adjustment):
    # 
    value = (math.sqrt(sel_value)*r_ness)*adjustment
    return value

######################################################################################################################################

def fn_set_metal_colour2(mc):
    # Returns mode for metallic colours and the colour name
    if (mc == 1):
        n_colour = "Gold"
        pdb.gimp_context_set_foreground(lightgold)
        pdb.gimp_context_set_background(darkgold)
        c_mode = 45#SOFTLIGHT new mode
    elif (mc == 2):
        n_colour = "Copper"
        pdb.gimp_context_set_foreground(darkcopper)
        pdb.gimp_context_set_background(lightcopper)
        c_mode = 47#GRAIN MERGE new mode
    elif (mc == 3):
        n_colour = "Brass"
        pdb.gimp_context_set_foreground(darkbrass)
        pdb.gimp_context_set_background(lightbrass)
        c_mode = 45#SOFTLIGHT new mode
    elif (mc == 4):
        n_colour = "Bronze"
        pdb.gimp_context_set_foreground(darkbronze)
        pdb.gimp_context_set_background(lightbronze)
        c_mode = 26#LCH Color new mode
        #c_mode = 45#SOFTLIGHT new mode
    else:
        n_colour = "Silver"
        pdb.gimp_context_set_foreground(lightgrey)
        pdb.gimp_context_set_background(midgrey)
        c_mode = 45#SOFTLIGHT new mode

    return c_mode, n_colour

######################################################################################################################################
	
def proc_messagebox(message):
    old_handler = pdb.gimp_message_get_handler()
    pdb.gimp_message_set_handler(2)
    pdb.gimp_message(str(message))
    pdb.gimp_message_set_handler(old_handler)
	
######################################################################################################################################
		
def proc_trim_away(img, layer, amount):
    #### Trim off pixels from a layer
    # set our feather settings
    our_feather = True
    our_feather_radius = 2
    #get feather settings
    feather = pdb.gimp_context_get_feather()
    feather_x, feather_y = pdb.gimp_context_get_feather_radius()
    # make sure that feather is set
    if not feather: pdb.gimp_context_set_feather(our_feather)
    # use our feather radius
    pdb.gimp_context_set_feather_radius(our_feather_radius, our_feather_radius)
    # select the contents of the layer
    pdb.gimp_image_select_item(img, 2, layer)
    # shrink the selection
    pdb.gimp_selection_shrink(img, amount)
    # invert selection
    pdb.gimp_selection_invert(img)
    # trim away pixels
    successful = pdb.gimp_edit_cut(layer)
    # deselect
    pdb.gimp_selection_none(img)
    # reset feather settings
    pdb.gimp_context_set_feather(feather)
    pdb.gimp_context_set_feather_radius(feather_x, feather_y)
	
    #### End of Proc_trim_away
	
######################################################################################################################################
	
def proc_text_bevel(image, text_layer, bevel_width, bevel_depth, feather, black, white, midgrey, bevel_mode):
    #### Create a bevel around text and return the new layer as bevel_layer
    #Rem Select text outline and save it
    pdb.gimp_image_select_item(image, 2, text_layer)
    saved_selection = pdb.gimp_selection_save(image)

    #### Rem Make a bumpmap layer ######
    bumpmap_layer = pdb.gimp_layer_new_from_drawable(text_layer, image)
    pdb.gimp_image_insert_layer(image, bumpmap_layer, None, -1)
    pdb.gimp_item_set_name(bumpmap_layer, "Text Layer Bumpmap")
    pdb.gimp_image_raise_item_to_top(image, bumpmap_layer)

    pdb.gimp_selection_feather(image, feather)
    blurred_selection = pdb.gimp_selection_save(image)
    loop = bevel_width
    pdb.gimp_selection_shrink(image, loop)

    while (pdb.gimp_selection_is_empty(image) == True ):
        loop = loop - 1
        pdb.gimp_image_select_item(image, 2, blurred_selection)
        pdb.gimp_selection_shrink(image, loop)
        
    bevel_width = min(bevel_width, loop)
    pdb.gimp_image_select_item(image, 2, blurred_selection)

    pdb.gimp_context_set_foreground(black)
    pdb.gimp_drawable_fill(bumpmap_layer, 0)

    loop = 1
    #Rem Temp dirty fill_value 
    fill_value = (255 / bevel_width)
    while (loop <= bevel_width):

        #Rem Using Temp fill_value until something better comes to mind
        pdb.gimp_image_select_item(image, 2, blurred_selection)
        pdb.gimp_selection_shrink(image, loop -1)
        greycolour = int(fill_value)
        grey = gimpcolor.RGB(greycolour, greycolour, greycolour)
        pdb.gimp_context_set_foreground(grey)

        pdb.gimp_edit_fill(bumpmap_layer, 0)

        #Rem Temp fill_value
        fill_value = fill_value + (255 / bevel_width)
        loop = loop + 1

    pdb.gimp_context_set_foreground(white)
    pdb.gimp_image_select_item(image, 2, blurred_selection)
    pdb.gimp_selection_shrink(image, bevel_width)

    if (pdb.gimp_selection_is_empty(image) == FALSE):
        pdb.gimp_edit_fill(bumpmap_layer, 0)

    pdb.gimp_selection_none(image)

    #### Make Bevel from bumpmap ####
    bevel_layer = pdb.gimp_layer_new_from_drawable(text_layer, image)
    pdb.gimp_image_insert_layer(image, bevel_layer, None, -1)
    pdb.gimp_item_set_name(bevel_layer, "Bevel")
    pdb.gimp_image_raise_item_to_top(image, bevel_layer)

    pdb.gimp_context_set_foreground(midgrey)
    pdb.gimp_drawable_fill(bevel_layer, 0)
	
    pdb.plug_in_bump_map(image, bevel_layer, bumpmap_layer, 135, 35, bevel_depth, 0, 0, 0, 0, True, False, 0)
    pdb.gimp_layer_set_mode(bevel_layer, bevel_mode)
    pdb.gimp_layer_set_opacity(bevel_layer, 80)
	
    #### Delete area outside the bevel ####
    pdb.gimp_image_select_item(image, 2, saved_selection)
    pdb.gimp_selection_invert(image)

    if (pdb.gimp_selection_is_empty(image) == False):
        pdb.gimp_edit_clear(bevel_layer)

    pdb.gimp_image_select_item(image, 2, saved_selection)
    pdb.gimp_selection_shrink(image, bevel_width)

    if (pdb.gimp_selection_is_empty(image) == False):
        pdb.gimp_edit_clear(bevel_layer)
	
    pdb.gimp_image_select_item(image, 2, saved_selection)
 
    pdb.gimp_selection_none(image)
    pdb.gimp_drawable_set_visible(bumpmap_layer, False)	
    pdb.gimp_image_remove_channel(image, saved_selection)
    pdb.gimp_image_remove_channel(image, blurred_selection)
    pdb.gimp_image_remove_layer(image, bumpmap_layer)
    pdb.gimp_image_raise_item_to_top(image, bevel_layer)
    pdb.gimp_item_set_linked(bevel_layer, True)
    return bevel_layer
	
#### End of proc_text_bevel

######################################################################################################################################
	
register(
    "python_fu_Custom_Font_Metallic_Effects_v0-10_Tabbed",
    "Provides options for applying metallic effects to Gimp Custom Fonts",
    "Provides options for applying metallic effects to Gimp Custom Fonts",
    "Skinnyhouse",
    "Open source (BSD 3-clause license)",
    "2022",
    "<Image>/Filters/Custom Font Tools/Custom Font Metallic Effects v0.10 Tabbed...",
    "RGB*",
    [
    (PF_OPTION, "pf_option_texture", "Metal texture: ", 1,("Glitter effect","Bright & shiny","Verdigris (can use optional texture layer)","Dulled","Annealed","Brushed horizontal", "Plated (better with wider fonts)"), {'newLine':'1','tab':0}),

    (PF_OPTION, "pf_option_colour", "Metallic Colour: ", 1,("Silver","Gold","Copper", "Brass", "Bronze"), {'newLine':'1','tab':0}),
    (PF_OPTION, "pf_option_curves", "Curves applied to plate texture: ", 5,("As Glitter", "As Bright","As Verdigris","As Dull","As Annealed", "Polish 1", "Polish 2", "Polish 3", "Polish 4", "Polish 5", "Polish 6", "Polish 7", "Polish 8"), {'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_plate", "Plate softness modifier: ", 5,(1,9,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_border", "Plate border modifier: ", 3,(1,9,1),{'newLine':'1','tab':0}),
    (PF_TOGGLE, "pf_usertextureoff",   "Employ original texture: ", 0, [("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_SLIDER, "pf_opacity", "Percentage of metal texture to use\nwith original texture: ", 50, (5, 100, 5)),
    (PF_TOGGLE, "pf_stroke_gradient", "Replace metallic with current gradient: ", 0,[("No"),("Yes") ],{'newLine':'1','tab':0}),
    (PF_GRADIENT, "pf_current_gradient", "Current gradient: ", 0,{'newLine':'1','tab':0}),
    (PF_OPTION, "mode_option2", "Metal colour mode: ", 26, ("Normal", "Dissolve","Color erase", "Merge", "Split", "Lighten only", "Luma/Luminance lighten only", "Screen", "Dodge", "Addition", "Darken only", "Luma/Luminance darken only", 
    "Multiply", "Burn", "Linear burn", "Overlay", "Soft light", "Hard light", "Vivid light", "Pin light", "Linear light", "Hard mix", "Difference", "Exclusion", "Subtract", "Grain extract", "Grain merge", 
    "Divide", "Hue (HSV)", "Saturation (HSV)", "Color (HSL)", "Value (HSL)", "Hue (LCH)", "Chroma (LCH)", "Color (LCH)", "Lightness (LCH)", "Luminance"),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_bump_depth", "Bumpmap depth (as percentage): ", 30,(0,100,1),{'newLine':'1','tab':0}),
    (PF_SPINNER, "pf_final_bump_depth", "Final bumpmap depth (as percentage): ", 0,(0,100,1),{'newLine':'1','tab':0}),
	
    (PF_SPINNER, "pf_light_amount", "Distance of light source\nfrom layer centre: ", 1.50,(0.00,1000.00,0.10),{'newLine':'1','tab':1}),
    (PF_OPTION, "pf_compass", "Point light position or\nlight direction as compass point: ", 7,("From North","From North East","From East","From South East","From South","From South West","From West","From North West"),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_effect_elevation", "Effect elevation: ", 30.0,(0.5,90,0.1),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_userandomsheenonoff",   "Randomize sheen: ", 0,[("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_userlightingoff",   "Apply lighting plug-in: ", 0,[("No"),("Yes") ],{'newLine':'1','tab':1}),
    (PF_SLIDER, "pf_shade_of_grey", "Percentage of black in shade of grey\napplied to base layer or original texture: ", 50, (10, 90, 5),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_blur_radius", "Artifact blur radius: ", 0.1,(0.0,50.0,0.1),{'newLine':'1','tab':1}),
    (PF_SPINNER, "pf_blur_delta", "Artifact blur delta amount: ", 10,(0,255,1),{'newLine':'1','tab':1}),
    (PF_TOGGLE, "pf_CF_off",   "Apply filter method: ", 1,[("To currently selected layer only"),("To custom font layers") ],{'newLine':'1','tab':1}),

    (PF_TAB    , "ptab"    , "tabs", 0, (_('Style, Texture, Colour'),_('Lighting, Blur & Method')) ),
    ],
    [],
    do_stuff)

main()
